#!/usr/bin/env ruby
#    Rubyripper - A secure ripper for Linux/BSD/OSX
#    Copyright (C) 2007  Bouke Woudstra (rubyripperdev@gmail.com)
#
#    This file is part of Rubyripper. Rubyripper is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>

#ensure translation is working before actually installing
LOCALE=[ENV['PWD'] + "/locale", "/usr/share/locale"]
LOCALE.each{|dir| if File.directory?(dir) : ENV['GETTEXT_PATH'] = dir ; break end}

$rr_version = '0.5.0' #application wide setting

begin
  require 'gettext'
rescue LoadError
  puts "RubyGems is not found. Please install ruby-gettext."
end

require 'monitor' #help library for threaded applications

def installed(filename) # a help function to check if an application is installed
	ENV['PATH'].split(':').each do |dir|
		if File.exist?(dir + '/' + filename) : return true end
	end
	if File.exist?(filename) : return true else return false end #it can also be in current working dir
end

if not installed('cdparanoia')
	puts "Cdparanoia not found on your system.\nThis is required to run rubyripper. Exiting..."
	exit()
end

def cdrom_drive #default values for cdrom drives under differenty os'es
	drive = 'Unknown!'
	system = RUBY_PLATFORM
	if system.include?('linux') || system.include?('bsd')
		drive = '/dev/cdrom'
	elsif system.include?('darwin')
		drive = '/dev/disk1'
	elsif system.include?('win') # no support for Windows as of yet, but keeping possibilities open for the future
		drive = 'cdaudio'
	end
	return drive
end

def get_example_filename_normal(basedir, layout) #separate function to make it faster
	filename = File.join(basedir, layout)
	filename = File.expand_path(filename)
	filename = "Example filename: #{filename}.ext"
	{'%a' => 'Judas Priest', '%b' => 'Sin After Sin', '%f' => 'codec', '%g' => 'Rock', '%y' => '1977', '%n' =>'01', '%t' => 'Sinner', '%i' =>'inputfile', '%o' => 'outputfile'}.each{|key, value| filename.gsub!(key,value)}
	return filename
end

def get_example_filename_various(basedir, layout) #separate function to make it faster
	filename = File.join(basedir, layout)
	filename = File.expand_path(filename)
	filename = "Example filename: #{filename}.ext"
	{'%a' => 'Various Artists', '%b' => 'TMF Rockzone', '%f' => 'codec', '%g' => "Rock", '%y' => '1999', '%n' => '01', '%va' => 'Kid Rock', '%t' => 'Cowboy'}.each{|key, value| filename.gsub!(key,value)}
	return filename
end

def get_filename(settings, codec = 'mp3', track = 1, command = false) #function returning filename after conversion of variables
	if settings['cd'].various_artists.empty?
	 	filename =  File.join(settings['basedir'], settings['naming_normal'])
	else
		filename = File.join(settings['basedir'], settings['naming_various'])
	end
	filename = File.expand_path(filename) #Replace any ~/ with the relevant information, such as /home/username/
	
	if command != false
		command.gsub!('%o', filename) #replace the %o with the output filename
		filename = command
	end
	
	if (filename.include?('%va') && settings['cd'].various_artists.empty?)
		filename.gsub!('%va', '')
		puts "Warning: you are using the various artist (%va) in the naming scheme for normal cd's!"
		puts "This is automatically removed"
	end
	{'%a' => settings['cd'].artist, '%b' => settings['cd'].album, '%f' => codec, '%g' => settings['cd'].genre, '%y' => settings['cd'].year, '%n' => sprintf("%02d", track), '%va' => settings['cd'].various_artists[track - 1], '%t' => settings['cd'].tracklist[track - 1]}.each{|key,value|  if filename.include?(key) : filename.gsub!(key, value) end}
	return filename
end

class Gui_support
attr_reader :update_ripping_progress, :update_encoding_progress, :append_2_log, :summary, :ripping_progress, :encoding_progress, :mismatch, :short_summary, :delete_logfiles
	include GetText
	bindtextdomain("rr_lib")

	def initialize(settings, dirnames) #gui is an instance of the graphical user interface used
		@settings = settings
		@logfiles = Array.new
		dirnames.each{|dir| @logfiles << File.open(dir + "/ripping.log", "a")} # Open a log file in append mode for each directory
		@problem_tracks = Hash.new # key = tracknumber, value = new dictionary with key = seconds_chunk, value = [amount_of_chunks, trials_needed]
		@not_corrected_tracks = Array.new # Array of tracks that weren't corrected within the maximum amount of trials set by the user
		@ripping_progress = 0.0
		@encoding_progress = 0.0
		@short_summary = ""
		update_logfiles(_("This log is created by Rubyripper, version %{version}\n") % {:version=>$rr_version})
		update_logfiles(_("Website: http://code.google.com/p/rubyripper\n\n"))
	end
	
	def update_ripping_progress(new_value, calling_function = false) #new_value = float, 1 = 100%
		new_value <= 1.0 ? @ripping_progress = new_value : @ripping_progress = 1.0
		@settings['instance'].update("ripping_progress", @ripping_progress)
	end
	
	def update_encoding_progress(new_value, calling_function = false) #new_value = float, 1 = 100%
		new_value <= 1.0 ? @encoding_progress = new_value : @encoding_progress = 1.0
		@settings['instance'].update("encoding_progress", @encoding_progress)
	end
	
	def append_2_log(message, calling_function = false)
		@logfiles.each{|logfile| logfile.print(message); logfile.flush()} # Append the messages to the logfiles
		@settings['instance'].update("log_change", message)
	end
	
	def update_logfiles(message, summary = false)
		@logfiles.each{|logfile| logfile.print(message); logfile.flush()} # Append the messages to the logfiles
		if summary : @short_summary += message end
	end
	
	def mismatch(track, trial, indexes_with_errors, size, length)
		if !@problem_tracks.key?(track) #First time we encounter this track (Secure_rip->analyzeFiles() )
			@problem_tracks[track] = Hash.new # create the Hash for the track
			indexes_with_errors.each do |index_of_chunk|
				seconds = index_of_chunk / 176400 # position of chunk rounded in seconds, each second = 176400 bytes
				if !@problem_tracks[track].key?(seconds)
					@problem_tracks[track][seconds] = [1, trial] # different_chunks, trial. First time we encounter this position, so different_chunks = 1
				else
					@problem_tracks[track][seconds][0] += 1 # one more chunk at the same second
				end
			end
		else
			indexes_with_errors.each do |index_of_chunk|
				seconds = index_of_chunk / 176400 # position of chunk rounded in seconds, each second = 176400 bytes
				@problem_tracks[track][seconds][1] = trial #Update the amount of trials needed
			end
		end
		if trial == 0: @not_corrected_tracks << track end #Reached maxtries and still got errors
	end
	
	def summary(matches_all, matches_errors, maxtries) #Give an overview of errors
		update_logfiles(_("\nRIPPING SUMMARY\n\n"), true)
		
		update_logfiles(_("All chunks were tried to match at least %{matches_all}  times.\n") % {:matches_all => matches_all}, true)
		if matches_all != matches_errors: update_logfiles(_("Chunks that differed after %{matches_all} trials,\nwere tried to match %{matches_errors} times.\n") %{:matches_all=>matches_all, :matches_errors=>matches_errors}, true) end

		if @problem_tracks.empty?
 			update_logfiles(_("None of the tracks gave any problems\n"), true)
 		elsif @not_corrected_tracks.size != 0
 			update_logfiles(_("Some track(s) could NOT be corrected within the maximum amount of trials\n"), true)
 			@not_corrected_tracks.each do |track|
				update_logfiles(_("Track %{track} could NOT be corrected completely\n") %{:track=>track}, true)
			end
 		else
 			update_logfiles(_("Some track(s) needed correction,but could\nbe corrected within the maximum amount of trials\n"), true)
 		end

 		if !@problem_tracks.empty? # At least some correction was necessary
			position_analyse(matches_errors, maxtries)
			@short_summary += _("The exact positions of the suspicious chunks\ncan be found in the ripping log\n")
		end
		@logfiles.each{|logfile| logfile.close} #close all the files
 	end
 		
 	def position_analyse(matches_errors, maxtries) # Give an overview of suspicion position in the logfile
		update_logfiles(_("\nSUSPICION POSITION ANALYSIS\n\n"))
		update_logfiles(_("Since there are 75 chunks per second, after making the notion of the\n"))
		update_logfiles(_("suspicion position, the amount of initially mismatched chunks for that position is shown.\n\n"))
		@problem_tracks.keys.sort.each do |track| # For each track show the position of the files, how many chunks of that position and amount of trials needed to solve
			update_logfiles(_("TRACK %{track}\n") %{:track=>track})
			@problem_tracks[track].keys.sort.each do |length| #length = total seconds of suspicious position
				minutes = length / 60 # ruby math -> 70 / 60 = 1 (how many times does 60 fit in 70)
				seconds = length % 60 # ruby math -> 70 % 60 = 10 (leftover)
				if @problem_tracks[track][length][1] != 0
					update_logfiles(_("\tSuspicion position : %{minutes}:%{seconds} (%{amount}x) (CORRECTED at trial %{trial}\n") % {:minutes=>sprintf("%02d", minutes), :seconds=>sprintf("%02d", seconds), :amount=>@problem_tracks[track][length][0], :trial=>@problem_tracks[track][length][1] + 1})
				else # Position could not be corrected
					update_logfiles(_("\tSuspicion position : %{minutes}:%{seconds} (%{amount}x) (COULD NOT BE CORRECTED)\n") %{:minutes => sprintf("%02d", minutes), :seconds=>sprintf("%02d", seconds), :amount=>@problem_tracks[track][length][0]})
				end
			end
		end
	end
	
	def delete_logfiles
		if @problem_tracks.empty? #only delete the logfile if no errors occured
			@logfiles.each{|logfile| File.delete(logfile.path)}
		end
	end
end

class Cuesheet # Started the very beginning
	def initialize(cddb, output_dirs, file)
		@cddb = cddb
		@output_dirs = output_dirs
		@cuesheet = ''
		@file = file
		toc = `cdparanoia -d #{@cdrom} 2>&1`.split("\n") #command
		
	end
	
	def analyze_toc # fill startpos, endpos, pregap
		@startpos = Array.new #Where does each track start?
		@endpos = Array.new #Where does each track end?
		@pregap = Array.new #Any space between tracks?
		
		while toc[0][0,2] != "==" : toc.shift end #remove all the non-interesting stuff
		toc.shift() #remove the ============ line
			
		@cddb.audiotracks.times do |time|
			temp = toc[time].split() # temp = [track, lengt_sector, length_time, start_sector]
			@startpos << temp[3].to_i # get the startpos
			@endpos << temp[1].to_i + temp[3].to_i #endpos = length + startpos
		end
		
		@startpos.size.times do |time| # Save pregap
			if time == 0: @pregap << @startpos[time]
			else @pregap << @startpos[time] - @endpos[time - 1]
			end
		end
	end
	
	def time(sector) # minutes:seconds:leftover frames
		minutes = sector / 4500 # 75 frames/second * 60 seconds/minute
		seconds = (sector % 4500) / 75
		frames = sector % 75 # leftover
		return "#{sprintf("%02d", minutes)}:#{sprintf("%02d", seconds)}:#{sprintf("%02d", frames)}"
	end
	
	def cuesheet_singlefile()
		@cuesheet += "REM GENRE #{@cddb.genre}\n"
		@cuesheet += "REM DATE #{@cddb.year}\n"
		@cuesheet += "COMMENT \"Rubyripper 0.3\"\n"
		@cuesheet += "PERFORMER \"#{@cddb.artist}\"\n"
		@cuesheet += "TITLE \"#{@cddb.album}\"\n"
		@cuesheet += "FILE \"#{@file}\" WAVE\n" #@file = filename
		@cddb.audiotracks.times do |track|
			@cuesheet += "\tTRACK #{sprintf("%02d", track + 1)} AUDIO\n"
			@cuesheet += "\t\tTITLE \"#{@cddb.tracklist[track]}\"\n"
			@cuesheet += "\t\tPERFORMER \"#{@cddb.artist}\"\n"
			if track > 0 && @pregap[track] > 0: @cuesheet += "\t\tINDEX 00 #{time(@endpos[track - 1])}\n" #Only if there is a pregap
			elsif track == 0 && @pregap[0] > 0: @cuesheet += "\t\tINDEX 00 00:00:00\n"
			end
			@cuesheet += "\t\tINDEX 01 #{time(@startpos[track])}\n"
		end
	end
end

class Cddb

attr_reader :audiotracks, :raw_response, :tracks_seconds, :make_internet_contact, :cdrom, :devicename, :totalsectors, :startsectors, :length_in_sectors, :tracks_playtime, :cd_playtime, :freedb, :freedb_choice, :default_drive, :update_yaml, :freedb_string, :restore_various
attr_accessor :artist, :album, :genre, :year, :tracklist, :various_artists

	include GetText
	bindtextdomain("rr_lib")
	
	def initialize(cdrom, instance=false, verbose=false, string=String.new)
		require 'yaml'
		@cdrom = cdrom
		@gui = instance
		@verbose = verbose
		@old_string = string #string of the previous disc
		@old_tracklist = Array.new #to restore various artists when detection is incorrect
		@various_artists = Array.new
		@freedb_string = String.new
		
		if installed('cdparanoia')
			scan_drive() #save the disc info generated by cdparanoia into @output
			parse_output() #get usefull variables from @output
			set_default_values() # default metadata
		else
			@gui.update("error", _("cdparanoia not found. Can't analyze CD!"))
		end
	end
	
	def scan_drive
		@default_drive = false # will be true if cdparanoia doesn't support the device parameter (-d)
		@output = `cdparanoia -d #{@cdrom} -vQ 2>&1` # fire a scan of the disc
		if @output.include?('USAGE')
			if @verbose : puts _("No disc found or cdparanoia doesn't support the drive parameter. Will retry with default drive") end
			@output = `cdparanoia -vQ 2>&1`
			if @output.include?("Unable to open disc") #No disc found by cdparanoia
				@gui.update("error", _("No audio disc found in %{cdrom}") % {:cdrom=>@cdrom}) 
			else 
				@default_drive = true # Apparently the -d parameter isn't supported. Value of default_drive will be used by Secure_RIp class
			end
		end
	end

	def parse_output # based on cdparanoia output
		@audiotracks = 0 ; 	@totalsectors = 0 ; @data_start = 0 ; @data_end = 0 #initialize values to 0, so we can later check if they're != 0
		@startsectors = Array.new ; @length_in_sectors = Array.new ; @tracks_playtime = Array.new ; @tracks_seconds = Array.new # arrays with values for each track

		@devicename = _("Unknown") ; @cd_playtime = '00:00'
		
		@output.split("\n").each do |line|
			if line[0,5] =~ /\s+\d+\./ #Bingo! We've got the info of a track s = space, d = digit, + means 1 or more
				trackinfo = line.split() # Split the values tracknumber, length in sectors, playtime, startsector, starttime
				@audiotracks += 1 # each line present one track
				@startsectors << trackinfo[3].to_i #save the integer value
				@length_in_sectors << trackinfo[1].to_i
				@tracks_playtime << trackinfo[2][1,5]
			elsif line =~ /CDROM\D*:/
				@devicename = $'.strip()
			elsif line[0,5] == "TOTAL"
				@totalsectors = line.split()[1].to_i
				@cd_playtime = line.split()[2][1,5]
			elsif line[0,15] == "track_num = 176"   # 176 = 0xb0 = end of data track??? # only MacOS version prints this
				@data_end = line.split()[5].to_i
			end
		end

		@length_in_sectors.each{|sectors| @tracks_seconds << sectors / 75} # Each second has 75 sectors
		if @data_end != 0   # detected a data track; assume last track == data track
			rawtoc = @output.find_all{|line| line[0,9] == "track_num"}
			@data_start = rawtoc.pop().split()[5].to_i
		end
	end
	
	def set_default_values # for file naming and metadata
		@artist = _('Unknown') ; @album = _('Unknown') ; @genre = _('Unknown'); @year = '0'
		@tracklist = Array.new # Titles of the tracks
		@audiotracks.times{|number| @tracklist << _("Track %{number}") % {:number=> number + 1}}# temporary tracklist
	end

	def gen_freedb_string # manual generate freedb strings! Thanks Scott for providing the code :-)
		checksum = 0
		ssectors = @startsectors
		ntracks = @audiotracks
		if @data_end == 0 # not a data disc (?) so behave normally
			tsectors = @totalsectors
		else
			ntracks += 1
			tsectors = @data_end
			ssectors << @data_start
		end

		ntracks_sectors = ntracks.to_s + " "
		ssectors.each do |sector|
			ntracks_sectors += (sector + 150).to_s + " " #append start sectors, adding mysterious 150 frame offset
			second = (sector + 150) / 75 #convert sectors -> seconds
			second.to_s.each_byte{|digit| checksum += digit.chr.to_i }  #sum digits of track _second_ offset
		end
		checksum %= 255
		hex_id =  checksum.to_s(16).rjust(2, '0') #cksum to hex string
		hex_id += (tsectors / 75).to_s(16).rjust(4, '0') #total playtime to hex string
                hex_id += ntracks.to_i.to_s(16).rjust(2, '0') #track count to hex string

		# freedb_string = hex id + number of tracks + each track sector + total sectors
		@freedb_string = hex_id + " " + ntracks_sectors + ((tsectors + 150) / 75).to_s
        end
	
	def freedb(freedb_options=false, choose=false) # based on cd-discid output
		@freedb_options = freedb_options ? freedb_options : {"site" => "http://freedb2.org:80/~cddb/cddb.cgi", "username" => "anonymous", "hostname" => "my_secret.com"} # if not set, use defaults
		@choose = choose # pass multiple freedb hits to gui

		if installed('discid') # check if we can use discid program
			if RUBY_PLATFORM.include?('darwin') : `diskutil unmount #{@cdrom}` end
			@freedb_string = `discid #{@cdrom}`
			if RUBY_PLATFORM.include?('darwin') : `diskutil mount #{@cdrom}` end
		elsif installed('cd-discid') # if no discid exists, try cd-discid
			@freedb_string = `cd-discid #{@cdrom}`
		else # if both do not exist generate it ourselves (less foolproof)
			@gui.update("error", _("discid or cd-discid not found. Trying fallback..."))
			gen_freedb_string()
			puts _("Using fallback... freedb string = %{freedb_string}" %{:freedb_string=>@freedb_string})
		end
		get_disc_info()
	end
		
	def get_disc_info
 		if @freedb_string != "" # audio-disc is found
			 # If the disc is loaded before, use this info. Don't do this if the disc is reloaded a second time.
			
			if File.exists?(yaml_file = File.join(ENV['HOME'], '.rubyripper/freedb.yaml'))
				@freedb_hash = YAML.load(File.open(yaml_file))
			else
				@freedb_hash = Hash.new
			end
			
			if @freedb_hash.has_key?(@freedb_string) && @freedb_string != @old_string # do we have the freedb string in our collection of saved queries?
				@raw_response = @freedb_hash[@freedb_string]
				@tracklist.clear() #Now fill it with the real tracknames
				cd_vars()
				@gui.update("cddb_hit", false) # Give the signal that we're finished
			elsif searchLocal && @freedb_string != @old_string
				@tracklist.clear()
				cd_vars()
				@gui.update("cddb_hit", false)
			else
				if @verbose : puts "preparing to contact freedb server" end
				make_internet_contact()
			end
 		else
 			@gui.update("error", _("No audio disc found in %{cdrom}") % {:cdrom=>@cdrom})
 		end
	end
			
	def searchLocal
		if File.directory?(dir = File.join(ENV['HOME'], '.cddb'))
			Dir.foreach(dir) do |subdir|
				if subdir == '.' || subdir == '..' : next end
				Dir.foreach(File.join(dir, subdir)) do |file|
					if file == @freedb_string[0,8]
						puts "Local file found #{File.join(dir, subdir, file)}"
						@raw_response = File.read(File.join(dir, subdir, file))
						return true
					end
				end
			end
		end
		return false
	end
	
	def make_internet_contact
		require 'net/http' #automatically loads the 'uri' library
		require 'cgi' #for communicating with the server
		require 'timeout' # in case internet doesn't work
		
		@url = URI.parse(@freedb_options['site'])
		@server = Net::HTTP.new(@url.host, @url.port)
		@query_string = @url.path + "?cmd=cddb+query+" + CGI.escape("#{@freedb_string.chomp}") + "&hello=" + CGI.escape("#{@freedb_options['username']} #{@freedb_options['hostname']} rubyripper #{$rr_version}") + "&proto=6"
		if @verbose : puts "Created query string: #{@query_string}" end
				
		begin # make contact now some pre-work is done
			Timeout::timeout(10) do
				respons, @answer = @server.get(@query_string)
			end
		rescue
			@gui.update("error", _("Couldn't fetch the freedb info within 10 seconds.\n\nDefault values will be shown..."))
			sleep(3) #give some time to read the message
			@gui.update("cddb_hit", false) # Give the signal that we're finished
			return false #quit this function
		end
			
		begin
			get_metadata()
		rescue
			puts "Exception thrown: #{$!}"
			@gui.update("error", _("Couldn't connect to freedb server. Network down?\n\nDefault values will be shown..."))
			sleep(3) #give some time to read the message
			@gui.update("cddb_hit", false) # Give the signal that we're finished
		end
	end	

	def get_metadata() # ask for matches on cd, if there are multiple, interaction with user is possible
		if @answer[0..2] == '200'  #There was only one hit found
			puts "One hit found; parsing"
			temp, @category, @discid = @answer.split()
			freedb_choice(0, true)
		elsif @answer[0..2] == '211' || @answer[0..2] == '210' #Multiple hits were found
			puts "Multiple hits found"
			get_choices()
			if (!@choose || @choices.length < 3) : freedb_choice(0) #Always choose the first one
			else @gui.update("cddb_hit", @choices) #Let the user choose
			end	
		elsif @answer[0..2] == '202'
			@gui.update("error", _("No match in Freedb database. Default values are used."))
			@gui.update("cddb_hit", false)
		else
			@gui.update("error", _("cddb_query return code = %{return_code}. Return code not supported.") % {:return_code=>@answer[0..2]})
			@gui.update("cddb_hit", false)
		end
	end
	
	def freedb_choice(choice, single=false) #if single=true there's no choise, otherwhise the choice is used for raw_response
		unless single # unless always first choice is used
			if choice == (@choices.length - 1) # keep defaults, don't get freedb (see get_choices function)
				@gui.update("cddb_hit", false)
				return true
			end
			@category, @discid = @choices[choice].split
		end
		get_raw_response()
		@tracklist.clear() #Now fill it with the real tracknames
		cd_vars()
		@gui.update("cddb_hit", false) # Give the signal that we're finished
	end

	def get_choices()	#Puts the multiple choices in an array
		@choices = Array.new
		answer_list = @answer.split("\n")[1..@answer.length]; # remove the first line, which we know is the header
		answer_list.each { | line | @choices << line.strip() unless line.strip() == "." }
		@choices << _("Keep defaults / don't use freedb") #also use the option to keep defaults
	end

	def get_raw_response #Retrieve all usefull metadata into @raw_response
		@fetch_string = @url.path + "?cmd=cddb+read+" + CGI.escape("#{@category} #{@discid}") + "&hello=" + CGI.escape("#{@freedb_options['username']} #{@freedb_options['hostname']} rubyripper #{$rr_version}") + "&proto=6"
		if @verbose : puts "Created fetch string: #{@query_string}" end
		
		respons, answer = @server.get(@fetch_string)
		answer_array = answer.split("\n")
		@raw_response = []
		answer_array.each do |line|
			line.chomp!
			@raw_response << line unless (line == nil || line[-1,1] == '=' ||line[0,1] == '#' || line[0,1] == '.' )
		end
		save_response()
	end
	
	def save_response(response = false) #now save the @raw_response(false) or the modified metadata (response=true)
		if not File.directory?(dirname = File.join(ENV['HOME'], '.rubyripper')) #check if dir exists
			Dir.mkdir(dirname)
		end

		if File.exists?(yaml_file = File.join(ENV['HOME'], '.rubyripper/freedb.yaml')) #check if file exists, if not then @freedb_hash is an empty hash
			@freedb_hash = YAML.load(File.open(yaml_file))
		end

		if response != false : @freedb_hash[@freedb_string] = response else @freedb_hash[@freedb_string] = @raw_response end
		yaml_file = File.new(yaml_file, 'w')
		yaml_file.write(@freedb_hash.to_yaml)
		yaml_file.close()
	end
	
	def update_yaml
		raw_response = String.new #now build a Freedb string
		raw_response += "DTITLE=#{@artist} \/ #{@album}\n"
		raw_response += "DYEAR=#{@year}\n"
		raw_response += "\nDGENRE=#{@genre}\n"
		if @various_artists.empty? : tracklist = @tracklist else tracklist = @old_tracklist end
		tracklist.each_index{|index| raw_response += "TTITLE#{index}=#{tracklist[index]}\n"}
		save_response(raw_response)
		clean_vars() #get any weird characters out of the filenames
	end
	
	def cd_vars #Make some usefull variables from the raw_response.
		@raw_response.each do |line|
			line.strip! #remove any newline characters
			if line =~ /DTITLE=/
				if @artist == _('Unknown') #first time we look at a DTITLE field (can have two lines at maximum)
					(@artist, @album) = $'.split(/\s*\/\s*/) # Remove the '/' with spaces around it, example DTITLE= Judas Priest     /     Sin After Sin
					if @artist == nil: @artist = _('Unknown') end
					if @album == nil: @album = _('Unknown') end
				elsif $' != nil # 2nd line with DTITLE, assume the second line is the continuation of the album name
					buf = $'
					@album = "#{@album}#{buf}"
				end
			elsif line =~ /DYEAR=/
				@year = $' ; if @year == nil : @year = 0 else @year end
			elsif line =~ /DGENRE=/
				@genre = $' ; if @genre == nil : @genre = _('Unknown') end
			elsif line =~ /TTITLE\d*=/
				trackname = $' # may also include variable artist
				line =~ /\d+/ ; tracknumber = $&.to_i #ruby magic, $& == the just matched regular expression
				if trackname != nil && @tracklist.empty? #1st track
					@tracklist << trackname
				elsif trackname != nil && @tracklist.length == tracknumber #counting of tracknumber starts with 0
					@tracklist << trackname 
				elsif trackname != nil #already had a line for this track, so add this trackname to the previous one
					@tracklist[-1] += trackname
				end
			end
		end
		check_for_various()
	end

#Various artist albums have different ways to show the artist and trackname
#Most common notation is ARTIST / TITLE
#Next to that is ARTIST - TITLE
#Then we got "TITLE" by ARTIST

	def check_for_various
		slash_sep = true ; hyphen_sep = true; by_sep = true
		@tracklist.each do |trackname| #walk through all tracks and see if in each track the separator is found
			unless trackname.include?('/')  : slash_sep = false end
			unless trackname.include?('-') : hyphen_sep = false end
			unless trackname.include?('by') : by_sep = false end
		end
		
		@old_tracklist = @tracklist.dup() #backup before overwrite with new values
		
		#if a separator is found, split the values accordingly, removing any spaces around the separator
		if slash_sep : @tracklist.each_index{|index| @various_artists[index], @tracklist[index] = @tracklist[index].split(/\s*\/\s*/)}
		elsif hyphen_sep : @tracklist.each_index{|index| @various_artists[index], @tracklist[index] = @tracklist[index].split(/\s*-\s*/)}
		elsif by_sep : @tracklist.each_index{|index| @tracklist[index], @various_artists[index] = @tracklist[index].split(/\s*by\s*/)}
		end
	end
	
	def restore_various
		@tracklist = @old_tracklist.dup
		@various_artists = Array.new
	end
	
	def clean_vars
		@album = clean(@album)
		@artist = clean(@artist)
		@year = clean(@year)
		@genre = clean(@genre)
		@tracklist.each_index{|index| @tracklist[index] = clean(@tracklist[index])}
		@various_artists.each_index{|index| @various_artists[index] = clean(@various_artists[index])}
	end

#The following characters are not allowed in Bash (in combination with other commands): the backquote( ` ), the double quote( " ), ( [ ), ( ] ), the forward slash ( / ), the backward slash ( \ )
#ruby passes by reference, functions with a '!' will immediately save the original value
	def clean(var)
		var.gsub!('`', "'") #replace a backquote with a single quote
		var.gsub!('"', '\"') #Add a slash before the double quote chars, otherwise the shell will complain
		var.gsub!('[', '(') #replace the ' [ ' with a ' ( '
		var.gsub!(']', ')') #replace the ' ] ' with a ' ) '
		var.gsub!('/', ' ') #remove any '/'
		var.gsub!('\\', '') #remove any '\', the \\ means a normal \
		var.gsub!('_', ' ') # replace any underscores with spaces, some freedb info got underscores instead of spaces
		var = var.strip
	end
end

class SecureRip
	include GetText
	bindtextdomain("rr_lib")

	def initialize(settings, encoding)
		@settings = settings
		@encoding = encoding
		@reqMatchesAll = @settings['req_matches_all'] # Matches needed for all chunks
		@reqMatchesErrors = @settings['req_matches_errors'] # Matches needed for chunks that didn't match immediately
		@progress = 0.0 #for the progressbar
		
		if @settings['maxThreads'] == 0 : ripTracks() else Thread.new{ripTracks()} end
	end
	
	def ripTracks
		@settings['log'].update_ripping_progress(0.0, "ripper") # Give a hint to the gui that ripping has started
		@settings['tracksToRip'].each do |track|
			puts "Ripping track #{track}" if @settings['debug']
		
			if @settings['offset'] != 0 && track == @settings['cd'].audiotracks && @settings['rippersettings'].include?('-Z') #workaround for bug in cdparanoia
				@settings['rippersettings'].gsub!(/-Z\s?/, '') #replace the -Z setting (and one space if it is there) with nothing. See issue nr. 13.
			end
			
			#reset next three variables for each track
			@errors = Hash.new()
			@filesizes = Array.new
			@trial = 0

			main(track)
			@encoding.addTrack(track) #ready to encode
		end
		
		if @settings['eject'] #eject the disc when finished if user checked this option
			Thread.new do
			 	if installed('eject') : `eject #{@settings['cd'].cdrom}`
				elsif installed('diskutil'): `diskutil eject #{@settings['cd'].cdrom}` #Mac users don't got eject, but diskutil
				else puts _("No eject utility found!")
				end
			end
		end
	end
	
	def main(track)
		@reqMatchesAll.times do # The amount of matches required tells how many times the files should be ripped before comparing
			@trial += 1
			rip(track)
			@filesizes << File.size("#{@settings['temp_dir']}track#{track}_#{@trial}.wav")
		end
		
		testFilesize(track) #Make sure that we got the needed files with the same size
		analyzeFiles(track) #If there are differences, save them in the @errors hash
				
		while @errors.size > 0
			if @trial > @settings['max_tries'] && @settings['max_tries'] != 0 # We would like to respect our users settings, wouldn't we?
				@settings['log'].append_2_log(_("Maximum tries reached. %{amount} chunk(s) didn't match the required %{amount2} times\n") % {:amount=>@errors.length, :amount2=>@reqMatchesErrors})
				@settings['log'].append_2_log(_("Will continue with the file we've got so far\n"))
				@settings['log'].mismatch(track, 0, @errors.keys, @sizeMatch, @settings['cd'].tracks_seconds[track - 1]) # zero means it is never solved.
				break # break out loop and continue using trial1
			end
			
			@trial += 1
			rip(track)
			
			if File.size("#{@settings['temp_dir']}track#{track}_#{@trial}.wav") == @sizeMatch # The file size must match with previous attempts
				if @trial > @reqMatchesErrors : correctErrorPos(track) else readErrorPos(track) end # If the reqMatches errors is equal of higher to @trial, no match would ever be found, so skip
			else
				@settings['log'].append_2_log(_("File size of trial %{trial} does not match previous attempts\n") %{:trial=> @trial})
			end
		end
		
		getDigest(track) # Get a MD5-digest for the logfile
		@progress += @settings['percentages'][track]
		@settings['log'].update_ripping_progress(@progress)
	end
	
	def testFilesize(track)
		@sizeMatch = false # Until proven otherwise :P
		if @filesizes.uniq.size == 1 : @sizeMatch = @filesizes[0] end # When there is no difference in the filesize proceed (this should happen in 99% of all cases)
		while !@sizeMatch
			@settings['log'].append_2_log(_("Still not %{amount} times the same filesize. Trying another time\n") % {:amount=>@reqMatchesAll})
			@trial += 1 ; rip() ; @filesizes << File.size("#{@settings['temp_dir']}track#{track}_#{@trial}.wav")
			filesizes = @filesizes.sort() # Make a sorted copy
			filesizes.each do |size| # This look like a lot of iteration, but most of the code is only run once
				if filesizes.rindex(size) - filesizes.index(size) >= (@reqMatchesAll - 1) # Do we have the required amount of same sizes?
					@sizeMatch = size
					trialnumber = 1 # So we have to iterate again with trialnumber as a help value.
					@filesizes.each_index do |index| # now clean up files with wrong filesizes. Since we sorted the copy, we don't know which trials have wrong filesizes.
						if @filesizes[index] == @sizeMatch 
							if File.exist?("#{@settings['temp_dir']}track#{track}_#{trialnumber}.wav") : trialnumber += 1 # The file won't exist if it is deleted earlier in the loop
							else File.rename("#{@settings['temp_dir']}track#{track}_#{index + 1}.wav", "#{@settings['temp_dir']}track#{track}_#{trialnumber}.wav") ; trialnumber += 1
							end
  						else
  							File.delete("#{@settings['temp_dir']}track#{track}_#{index + 1}.wav") # Delete files with non-matching filesizes
  						end
  					end
  					break # Since we have a filesize match, don't finish the loop
				end
			end
		end
	end
	
	def analyzeFiles(track)
		@settings['log'].append_2_log(_("Analyzing files for mismatching chunks\n"))
		files = Array.new
		@reqMatchesAll.times do |time|
			files << File.new("#{@settings['temp_dir']}track#{track}_#{time + 1}.wav", 'r')
		end
				
		(@reqMatchesAll - 1).times do |time|
			index = 0 ; files.each{|file| file.pos = 44} # 44 = wav container overhead, 2352 = size for a audiocd sector as used in cdparanoia
			while index + 44 < @sizeMatch
				if !@errors.key?(index) && files[0].read(2352) != files[time + 1].read(2352) # Does this sector matches the previous ones? and isn't the position already known?
					files.each{|file| file.pos = index + 44} # Reset each read position of the files
					@errors[index] = Array.new
					files.each{|file| @errors[index] << file.read(2352)} # Save the chunk for all files in the just created array
				end
				index += 2352
			end
		end
		
		files.each{|file| file.close}
		(@reqMatchesAll - 1).times{|time| File.delete("#{@settings['temp_dir']}track#{track}_#{time + 2}.wav")} # Don't need these files anymore. Differences are saved in memory.
		if @errors.size == 0
			@settings['log'].append_2_log(_("Every chunk matched %{amount} times :)\n") % {:amount=>@reqMatchesAll})
		else
			@settings['log'].mismatch(track, @trial, @errors.keys, @sizeMatch, @settings['cd'].tracks_seconds[track-1]) # report for later position analysis
			@settings['log'].append_2_log(_("%{amount} chunk(s) didn't match %{amount2} times.\n") % {:amount=>@errors.length, :amount2=>@reqMatchesAll})
		end
	end
	
	def readErrorPos(track) # This is used if the required amount of matches is too big to ever be realized
		file = File.new("#{@settings['temp_dir']}track#{track}_#{@trial}.wav", 'r')
		@errors.keys.sort.each do |start_chunk|
			file.pos = start_chunk + 44 # 44 = wav container overhead, 2352 = size for a audiocd sector as used in cdparanoia
			@errors[start_chunk] << file.read(2352)
		end
		file.close
		File.delete("#{@settings['temp_dir']}track#{track}_#{@trial}.wav") # Don't need this file anymore. Differences are saved in memory.
		@settings['log'].mismatch(track, @trial, @errors.keys, @sizeMatch, @settings['cd'].tracks_seconds[track-1]) # report for later position analysis
	end
	
	def correctErrorPos(track)
	#Let the errors 'wave' out. Chunks with different outcomes must be matches @reqMatchesErrors times.
		file1 = File.new("#{@settings['temp_dir']}track#{track}_1.wav", 'r+')
		file2 = File.new("#{@settings['temp_dir']}track#{track}_#{@trial}.wav", 'r')
		@errors.keys.sort.each do |start_chunk| # Sort the hash keys to prevent jumping forward and backwards in the file
			file2.pos = start_chunk + 44 # 44 = wav container overhead, 2352 = size for a audiocd sector as used in cdparanoia
			@errors[start_chunk] << temp = file2.read(2352)
			@errors[start_chunk].sort! # Sort the array before looking if temp solves our requirement of matches
			if (@errors[start_chunk].rindex(temp) - @errors[start_chunk].index(temp)) == (@reqMatchesErrors - 1) # Matches are r(ight)index - index because array is sorted
				file1.pos = start_chunk
				file1.write(temp)
				@errors.delete(start_chunk)
			end
		end
		file1.close
		file2.close
		File.delete("#{@settings['temp_dir']}track#{track}_#{@trial}.wav") # Don't need this file anymore. Differences are saved in memory.
		
		if @errors.size == 0
			@settings['log'].append_2_log(_("Error(s) succesfully corrected, %{amount} matches found for each chunk :)\n") % {:amount=>@reqMatchesErrors})
		else
			@settings['log'].mismatch(track, @trial, @errors.keys, @sizeMatch, @settings['cd'].tracks_seconds[track-1]) # report for later position analysis
			@settings['log'].append_2_log(_("%{amount} chunk(s) didn't match %{amount2} times.\n") % {:amount=>@errors.length, :amount2=>@reqMatchesErrors})
		end
	end
	
	def rip(track) # set cdparanoia command + parameters
		@settings['log'].append_2_log(_("Starting to rip track %{track}, trial %{trial}\#\n") % {:track=>track, :trial=>@trial})
		command = "cdparanoia"
		if @settings['rippersettings'].size != 0 : command += " #{@settings['rippersettings']}" end 
		command += (" #{track}") # second parameter tracknumber
		unless @settings['cd'].default_drive : command += " -d #{@settings['cdrom']}" end # the ported cdparanoia for MacOS misses the -d option, default drive will be used.
		command += " -O #{@settings['offset']}"
		command += " \"#{@settings['temp_dir']}track#{track}_#{@trial}.wav\"" #output name
		unless @settings['verbose'] : command += " 2>&1" end # hide the output of cdparanoia output
		`#{command}` #Launch the cdparanoia command
	end
	
	def getDigest(track)
		digest = Digest::MD5.new()
		file = File.open("#{@settings['temp_dir']}track#{track}_1.wav", 'r')
		index = 0
		while (index < @sizeMatch)
			digest << file.read(100000)
			index += 100000
		end
		file.close()
		@settings['log'].append_2_log(_("MD5 sum: %{md5sum}\n\n") %{:md5sum=>digest.hexdigest})
	end
end

class Encode < Monitor
	include GetText
	bindtextdomain("rr_lib")
	
	attr_reader :addTrack
	
	def initialize(settings)
		super()
		@settings = settings
		@progress = 0.0
		@threads = 0 # number of running threads
		@ready = true # is the encoding process idle or are we running out of threads?
		@waitingroom = Array.new # Room where each track and codec is listed, waiting for a free thread.
		@busyTracks = Array.new # Tracks that are currently encoding
		@lasttrack = false

		ENV['CHARSET'] = "UTF-8" #Needed for correct character handling for the oggenc executable. Perhaps other encoders use it as well.
		@codecs = 0 # number of codecs
		['flac','vorbis','mp3','wav','other'].each{|codec| if @settings[codec] : @codecs +=1 end}
	end
	
	def addTrack(track)
		['flac', 'vorbis', 'mp3', 'wav', 'other'].each do |codec|
			if @settings[codec]
				@waitingroom << [track, codec]
				puts "Adding track #{track} (#{codec}) into the waitingroom.." if @settings['debug']
			end
		end
		
		if @settings['normalize'] == 'normalize' && @settings['gain'] == 'album' && @settings['tracksToRip'][-1] != track
			return false # normalize in album mode prevents encoding untill the last wav file is ripped
		elsif @settings['normalize'] == 'normalize' && @settings['gain'] == 'album' && @settings['tracksToRip'][-1] == track
			normalize() # last track has been ripped, so normalize and continue encoding
		elsif @settings['normalize'] == 'normalize' && @settings['gain'] == 'track'
			normalize(track)
		end
		
		if track == @settings['tracksToRip'][0] : @settings['log'].update_encoding_progress(0.0) end # set it to 0% for the first track, so the gui shows it's started.
		if track == @settings['tracksToRip'][-1] : @lasttrack = true end
		if @ready == true : handleThreads() end
	end
	
	def handleThreads
		synchronize do #we don't want to mess up with shared variables, so synchronize
			while (@settings['maxThreads'] > @threads || @settings['maxThreads'] == 0) && @waitingroom.empty? == false
				puts "Inside handleThreads: maxthreads = #{@settings['maxThreads']}, threads = #{@threads}" if @settings['debug']	
				@threads += 1
				track, codec = @waitingroom.shift
				@busyTracks << track
				if @settings['maxThreads'] == 0
					encodeTrack(track, codec)
				else
					thread = Thread.new{ encodeTrack(track, codec) }
					thread.priority = -1 #Encoding gets low priority
				end
			end
			if @settings['maxThreads'] > @threads || @settings['maxThreads'] == 0 : @ready = true else @ready = false end
		end
	end
	
	def encodeTrack(track, codec)
		if codec == 'flac' : doFlac(track)
		elsif codec == 'vorbis' : doVorbis(track)
		elsif codec == 'mp3' : doMp3(track)
		elsif codec == 'wav' : doWav(track)
		elsif codec == 'other' : doOther(track)
		end
		
		if @settings['debug'] : puts "busytracks = #{@busyTracks}, track = #{track}, codec = #{codec}" end
		
		synchronize do #we don't want to mess up with shared variables, so synchronize
			@busyTracks.delete_at(@busyTracks.index(track)) #the wav source file is no longer needed for this codec, so delete the first mention of this track,
		
			if @waitingroom.flatten.include?(track) == false && @busyTracks.include?(track) == false && File.exists?("#{@settings['temp_dir']}track#{track}_1.wav")
				File.delete("#{@settings['temp_dir']}track#{track}_1.wav") #We don't need the wav file after it's encoded.
			end
	
			@threads -= 1
			@progress += @settings['percentages'][track] / @codecs
			@settings['log'].update_encoding_progress(@progress)
		end
		if @waitingroom.empty? && @threads == 0 && @lasttrack == true  : finished() ; return true end
		if @ready == false : handleThreads end #when a process is finished the loop for using all threads can restart, block all others meanwhile
	end
	
	def finished
		puts "Inside the finished function" if @settings['debug']
		@progress = 1.0 ; @settings['log'].update_encoding_progress(@progress)
		@settings['log'].summary(@settings['req_matches_all'], @settings['req_matches_errors'], @settings['max_tries'])
		if @settings['no_log']  : @settings['log'].delete_logfiles end #Delete the logfile if no correction was needed if no_log is true
		createPlaylists() if @settings['playlist']
		cleanup()
		@settings['instance'].update("finished")
	end
	
	def createPlaylists
		['flac', 'vorbis', 'mp3', 'wav', 'other'].each do |codec|
			if @settings[codec]
				dirname = File.dirname(get_filename(@settings,  codec, 1)) #tracknumber 1 for instance
				m3ufile = File.new(File.join(dirname, "#{@settings['cd'].artist} - #{@settings['cd'].album} (#{codec}).m3u"), 'w')
				Dir.entries(dirname).sort.each do |file|
					if File.file?(File.join(dirname, file)) && file != 'ripping.log' && file[-4,4] != '.m3u'
						m3ufile.puts file
					end
				end
				m3ufile.close()
			end
		end
	end
	
	def cleanup
		if File.directory?(@settings['temp_dir'])
			Dir.foreach(@settings['temp_dir']) do |file|
				if File.file?(filename = File.join(@settings['temp_dir'], file)): File.delete(filename) end
			end
			Dir.rmdir(@settings['temp_dir'])
		end
	end
	
	def normalize(track = false)
		if not installed('normalize') : puts "WARNING: Normalize is not installed. Cannot normalize files"; return false end
		
		if track == false #album mode
			command = "normalize -b \"#{@settings['temp_dir']}/\"*.wav"
			`#{command}`
		else
			command = "normalize \"#{@settings['temp_dir']}track#{track}_1.wav\""
			`#{command}`
		end
	end
	
	def replaygain(filename, codec, track)
		if @settings['normalize'] == "replaygain"
			if @settings['gain'] == "album" && @settings['tracksToRip'][-1] == track ||@settings['gain']=="track"
				if codec == 'flac'
					if not installed('metaflac') : puts "WARNING: Metaflac is not installed. Cannot replaygain files." ; return false end
					command = "metaflac --add-replay-gain \"#{if @settings['gain'] =="track" : filename else File.dirname(filename) + "\"/*.flac" end}"
					`#{command}`
				elsif codec == 'vorbis'
					if not installed('vorbisgain') : puts "WARNING: Vorbisgain is not installed. Cannot replaygain files." ; return false end
					command = "vorbisgain #{if @settings['gain'] =="track" : "\"" + filename + "\"" else "-a \"" + File.dirname(filename) + "\"/*.ogg" end}"
					`#{command}`
				elsif codec == 'mp3'
					if not installed('mp3gain') : puts "WARNING: Mp3gain is not installed. Cannot replaygain files." ; return false end
					command = "mp3gain -c #{if @settings['gain'] =="track" : "-r \"" + filename + "\"" else "-a \"" + File.dirname(filename) + "\"/*.mp3" end}"
					`#{command}`
				elsif codec == 'wav'
					if not installed('wavegain') : puts "WARNING: Wavegain is not installed. Cannot replaygain files." ; return false end
					command = "wavegain #{if @settings['gain'] =="track" : "\"" + filename +"\"" else "-a \"" + File.dirname(filename) + "\"/*.wav" end}"
					`#{command}`
				end
			end
		end
	end

	def doFlac(track)
		filename = get_filename(@settings,  'flac', track) + '.flac'
		if !@settings['flacsettings'] : @settings['flacsettings'] = '--best' end
		flac(filename, track)
		replaygain(filename, 'flac', track)
	end
		
	def doVorbis(track)
		filename = get_filename(@settings, 'vorbis', track) + '.ogg'
		if !@settings['vorbissettings'] : @settings['vorbissettings'] = '-q 6' end
		vorbis(filename, track)
		replaygain(filename, 'vorbis', track)
	end
		
	def doMp3(track)
		@possible_lame_tags = ['A CAPPELLA', 'ACID', 'ACID JAZZ', 'ACID PUNK', 'ACOUSTIC', 'ALTERNATIVE', 'ALT. ROCK', 'AMBIENT', 'ANIME', 'AVANTGARDE', \
'BALLAD', 'BASS', 'BEAT', 'BEBOB', 'BIG BAND', 'BLACK METAL', 'BLUEGRASS', 'BLUES', 'BOOTY BASS', 'BRITPOP', 'CABARET', 'CELTIC', 'CHAMBER MUSIC', 'CHANSON', \
'CHORUS', 'CHRISTIAN GANGSTA RAP', 'CHRISTIAN RAP', 'CHRISTIAN ROCK', 'CLASSICAL', 'CLASSIC ROCK', 'CLUB', 'CLUB-HOUSE', 'COMEDY', 'CONTEMPORARY CHRISTIAN', \
'COUNTRY', 'CROSSOVER', 'CULT', 'DANCE', 'DANCE HALL', 'DARKWAVE', 'DEATH METAL', 'DISCO', 'DREAM', 'DRUM & BASS', 'DRUM SOLO', 'DUET', 'EASY LISTENING', \
'ELECTRONIC', 'ETHNIC', 'EURODANCE', 'EURO-HOUSE', 'EURO-TECHNO', 'FAST-FUSION', 'FOLK', 'FOLKLORE', 'FOLK/ROCK', 'FREESTYLE', 'FUNK', 'FUSION', 'GAME', \
'GANGSTA RAP', 'GOA', 'GOSPEL', 'GOTHIC', 'GOTHIC ROCK', 'GRUNGE', 'HARDCORE', 'HARD ROCK', 'HEAVY METAL', 'HIP-HOP', 'HOUSE', 'HUMOUR', 'INDIE', 'INDUSTRIAL', \
'INSTRUMENTAL', 'INSTRUMENTAL POP', 'INSTRUMENTAL ROCK', 'JAZZ', 'JAZZ+FUNK', 'JPOP', 'JUNGLE', 'LATIN', 'LO-FI', 'MEDITATIVE', 'MERENGUE', 'METAL', 'MUSICAL', \
'NATIONAL FOLK', 'NATIVE AMERICAN', 'NEGERPUNK', 'NEW AGE', 'NEW WAVE', 'NOISE', 'OLDIES', 'OPERA', 'OTHER', 'POLKA', 'POLSK PUNK', 'POP', 'POP-FOLK', 'POP/FUNK', \
'PORN GROOVE', 'POWER BALLAD', 'PRANKS', 'PRIMUS', 'PROGRESSIVE ROCK', 'PSYCHEDELIC', 'PSYCHEDELIC ROCK', 'PUNK', 'PUNK ROCK', 'RAP', 'RAVE', 'R&B', 'REGGAE', \
'RETRO', 'REVIVAL', 'RHYTHMIC SOUL', 'ROCK', 'ROCK & ROLL', 'SALSA', 'SAMBA', 'SATIRE', 'SHOWTUNES', 'SKA', 'SLOW JAM', 'SLOW ROCK', 'SONATA', 'SOUL', 'SOUND CLIP', \
'SOUNDTRACK', 'SOUTHERN ROCK', 'SPACE', 'SPEECH', 'SWING', 'SYMPHONIC ROCK', 'SYMPHONY', 'SYNTHPOP', 'TANGO', 'TECHNO', 'TECHNO-INDUSTRIAL', 'TERROR', 'THRASH METAL', \
'TOP 40', 'TRAILER', 'TRANCE', 'TRIBAL', 'TRIP-HOP', 'VOCAL']
		filename = get_filename(@settings, 'mp3', track) + '.mp3'
		if !@settings['mp3settings'] : @settings['mp3settings'] = "--preset fast standard" end
		if @possible_lame_tags.include?(@settings['cd'].genre.upcase): genre = @settings['cd'].genre else genre = 'other' end
		mp3(filename, genre, track)
		replaygain(filename, 'mp3', track)
	end
		
	def doWav(track)
		filename = get_filename(@settings, 'wav', track) + '.wav'
		wav(filename, track)
		replaygain(filename, 'wav', track)
	end
	
	def doOther(track)
		command = get_filename(@settings, 'other', track, @settings['othersettings'].dup) #pass the commandline for other and replace all % fields (except %i)
		command.gsub!('%i', " \"#{@settings['temp_dir']}track#{track}_1.wav\"") # %i = input filename
		
		if @settings['debug'] : puts command end
		exec = IO.popen(command) #execute command
		if @settings['debug']: puts "Check exitstatus track #{track}, pid = #{exec.pid}" end
		if Process.waitpid2(exec.pid)[1].exitstatus != 0 : @settings['log'].append_2_log(_("WARNING: The 'other' codec exited with an error encoding track %{track}!\n") %{:track=>track}) end
	end
	
	def flac(filename, track)
		command = %Q{flac #{@settings['flacsettings']} -o "#{filename}" --tag ARTIST="#{@settings['cd'].various_artists.empty? ? @settings['cd'].artist : @settings['cd'].various_artists[track-1]}" --tag TITLE="#{@settings['cd'].tracklist[track-1]}" --tag ALBUM="#{@settings['cd'].album}" --tag DATE="#{@settings['cd'].year}" --tag TRACKNUMBER="#{track}" --tag GENRE="#{@settings['cd'].genre}" "#{@settings['temp_dir']}track#{track}_1.wav"#{" 2>&1" unless @settings['verbose']}}
		if @settings['debug'] : puts "executing command #{command}" end
		exec = IO.popen(command) #execute command
		if @settings['debug']: puts "Check exitstatus track #{track}, pid = #{exec.pid}" end
		if Process.waitpid2(exec.pid)[1].exitstatus != 0 : @settings['log'].append_2_log(_("WARNING: Flac exited with an error encoding track %{track}!\n") %{:track=>track}) end
	end
	
	def vorbis(filename, track)
		command = %Q{oggenc -o "#{filename}" #{@settings['vorbissettings']} -c DATE="#{@settings['cd'].year}" -c TRACKNUMBER="#{track}" -c TITLE="#{@settings['cd'].tracklist[track-1]}" -c ALBUM="#{@settings['cd'].album}" -c ARTIST="#{@settings['cd'].various_artists.empty? ? @settings['cd'].artist : @settings['cd'].various_artists[track-1]}" -c GENRE="#{@settings['cd'].genre}" "#{@settings['temp_dir']}track#{track}_1.wav" #{" 2>&1" unless @settings['verbose']}}
		exec = IO.popen(command) #execute command
		exec.readlines() #get all the output
		if @settings['debug']: puts "Check exitstatus track #{track}, pid = #{exec.pid}" end
		if Process.waitpid2(exec.pid)[1].exitstatus != 0 : @settings['log'].append_2_log(_("WARNING: Oggenc (vorbis) exited with an error encoding track %{track}!\n") %{:track=>track}) end
	end
	
	def mp3(filename, genre, track)
		command = %Q{lame --id3v2-only #{@settings['mp3settings']} --tt "#{@settings['cd'].tracklist[track-1]}" --ta "#{@settings['cd'].various_artists.empty? ? @settings['cd'].artist : @settings['cd'].various_artists[track-1]}" --tl "#{@settings['cd'].album}" --ty "#{@settings['cd'].year}" --tn "#{track}" --tg "#{genre}" "%i" "%o" #{" 2>&1" unless @settings['verbose']}}
		
		require 'iconv'
		command = Iconv.conv("ISO-8859-1", "UTF-8", command) #translate the UTF-8 string to latin. This is needed because of a lame bug.
		command.sub!('%o', filename) #the %o output file should stay in UTF-8
		command.sub!('%i', "#{@settings['temp_dir']}track#{track}_1.wav") #the %i input file should stay in UTF-8 as well
		exec = IO.popen(command) #execute command
		exec.readlines() #get all the output
		if @settings['debug']: puts "Check exitstatus track #{track}, pid = #{exec.pid}" end
		if Process.waitpid2(exec.pid)[1].exitstatus != 0 : @settings['log'].append_2_log(_("WARNING: Lame (mp3) exited with an error encoding track %{track}!\n") %{:track=>track}) end
	end
	
	def wav(filename, track)
		require 'fileutils'
		FileUtils.cp("#{@settings['temp_dir']}track#{track}_1.wav", filename)
	end
end

class Rubyripper
attr_reader :settings_ok, :start_rip, :output_dirs, :gui,  :postfix_dir, :overwrite_dir
	include GetText
	bindtextdomain("rr_lib")
	
	def initialize(settings)
		@settings = settings.dup
		@directory = false
		@settings['log'] = false
	end
	
	def settings_ok
	# update_yaml saves all the (changed?) metadata on the local disk and updates the different metadata fields in case of weird characters
		if check_configuration && test_for_deps && @settings['cd'].update_yaml()  && prepare_dirs
			return true
		else
			return false
		end
	end
	
	def start_rip
		update_gui() # Give some info about the cdrom-player, the codecs, the ripper, cddb_info
		compute_percentage() # Do some pre-work to get the progress updater working later on
		temp_dir() # Check if the temporary dir exists, if not: make it if basedir does exist or fallback to /tmp/rr/.
		require 'digest/md5' # Needed for secure class, only have to load them ones here.
		@encoding = Encode.new(@settings) #Create an instance for encoding
		@ripping = SecureRip.new(@settings, @encoding) #create an instance for ripping
	end
	
	def check_configuration
		unless File.symlink?(@settings['cdrom']) || File.blockdev?(@settings['cdrom']) # does the cdrom device or a symlink to it exist?
			@settings['instance'].update("error", _("The device %{cdrom} doesn't exist on your system!") %{:cdrom=>@settings['cdrom']})
			return false
		end

		if @settings['tracksToRip'].size == 0
			@settings['instance'].update("error", _("Please select at least one track."))
			return false
		end
		
		temp = Cddb.new(@settings['cdrom'], @settings['instance'])
		if @settings['cd'].cd_playtime != temp.cd_playtime # Test if the user has put in another CD to fool me ;)
			@settings['instance'].update("error", _("The Gui doesn't match inserted cd. Please press Scan Drive first."))
 			return false
		end
		
		unless @settings['flac'] || @settings['vorbis'] || @settings['mp3'] || @settings['wav'] || @settings['other'] # Test if at least one codec is selected
			@settings['instance'].update("error", _("No codecs are selected!"))
			return false
 		end
 		
		#Make sure errors get at least as much corrected as non-errors.
 		if @settings['req_matches_all'] > @settings['req_matches_errors'] : @settings['req_matches_errors'] = @settings['req_matches_all'] end
		return true
	end
	
	def test_for_deps()
		{"ripper" => "cdparanoia", "flac" => "flac", "vorbis" => "oggenc", "mp3" => "lame"}.each do |setting, binary|
			if @settings[setting] && !installed(binary)
				@settings['instance'].update("error", _("%{codec} not found on your system!") %{:codec=>binary.capitalize})
				return false
			end
		end
		return true
	end
	
	def prepare_dirs() # Look if the dirs already exist and create them
		if @settings['cd'].various_artists.empty? # if no various artist cd is found
			destination_dir = File.join(@settings['basedir'], @settings['naming_normal'])
		else
			destination_dir = File.join(@settings['basedir'], @settings['naming_various'])
		end
			
		if destination_dir.include?("/%b") && @settings['cd'].album[0,1] == '.' #If a dir or file starts with an album and the album starts with a dot 
			@settings['cd'].album.sub!(/\.*/, '') #remove those dot(s) from the albumname, otherwise they're hidden files in linux.
		end
		
		require 'ftools' # Adding some extra File functions like makedirs
	
		@output_dirs = Array.new
		['flac', 'vorbis', 'mp3', 'wav', 'other'].each do |codec| #save the dirname in @output_dirs
			if @settings[codec] #for each selected codec, save the dirname
				@output_dirs << File.dirname(get_filename(@settings, codec))
				unless destination_dir.include?('%f') : break end #only run the code ones if no codec format is found, otherwise the dirs are the same
			end
		end

		@output_dirs.each do |dir| #look for each dir if it exists
			 if File.directory?(dir)
			 	@settings['instance'].update("dir_exists", dir)
				return false 
			 end
		end
		
		@output_dirs.each do |dir| #Look for the highest existing dir and check if we have writing access there.
			while not File.directory?(dir)
				dir = File.dirname(dir) #chop of the subdir if it's not yet created
			end
			if not File.writable?(dir)
				@settings['instance'].update("error", _("Can't create output directory!\nYou have no writing acces in dir %{dir}") %{:dir=>dir})
				return false
			end
		end
		
		@output_dirs.each{|dir| File.makedirs(dir)} #create the dir
		
		@settings['log'] = @gui = Gui_support.new(@settings, @output_dirs) # The gui communication class :)
		return true
	end
	
	def postfix_dir #add the first free number as a postfix to the output dir
		postfix = 1
		@output_dirs.each{|dir| while File.directory?(dir + "\##{postfix}") : postfix += 1 end} #increase enough for all directories

		if @settings['cd'].various_artists.empty? #use a trick with File.join to remove any trailing '/'
			@settings['naming_normal'] = File.join(File.dirname(@settings['naming_normal']) + "\##{postfix}", File.basename(@settings['naming_normal']))
		else
			@settings['naming_various'] = File.join(File.dirname(@settings['naming_various']) + "\##{postfix}", File.basename(@settings['naming_various']))
		end
		prepare_dirs() #problem is solved, so continue
	end
	
	def overwrite_dir #overwrite the dir
		@output_dirs.each do |dir|
			if File.directory?(dir)
				Dir.foreach(dir){|file| if File.file?(File.join(dir,file)) : File.delete(File.join(dir, file)) end} #first delete all files in the dir
				Dir.rmdir(dir) #then delete the dir itself
			end
		end
		prepare_dirs() #problem is solved, so continue
	end

	def update_gui
		@settings['log'].append_2_log(_("Cdrom player used to rip:\n%{drivename}\n") %{:drivename=>@settings['cd'].devicename})
		@settings['log'].append_2_log(_("Cdrom offset used: %{offset}\n\n") %{:offset=>@settings['offset']})
		@settings['log'].append_2_log(_("Ripper used: cdparanoia %{settings}\n") % {:settings=> if @settings['rippersettings']: @settings['rippersettings'] else _('default settings') end})
		@settings['log'].append_2_log(_("Matches required for all chunks: %{amount}\n") %{:amount=>@settings['req_matches_all']})
		@settings['log'].append_2_log(_("Matches required for erroneous chunks: %{amount}\n\n") % {:amount=>@settings['req_matches_errors']})
		
		@settings['log'].append_2_log(_("Codec(s) used:\n"))
		if @settings['flac']: @settings['log'].append_2_log(_("-flac \t-> %{settings} (%{version})\n") %{:settings=>@settings['flacsettings'], :version=>`flac --version`.strip}) end
		if @settings['vorbis']: @settings['log'].append_2_log(_("-vorbis\t-> %{settings} (%{version})\n") %{:settings=>@settings['vorbissettings'], :version=>`oggenc --version`.strip}) end
		if @settings['mp3']: @settings['log'].append_2_log(_("-mp3\t-> %{settings}\n(%{version}\n") %{:settings=>@settings['mp3settings'], :version=>`lame --version`.split("\n")[0]}) end
		if @settings['wav']: @settings['log'].append_2_log(_("-wav\n")) end
		if @settings['other'] : @settings['log'].append_2_log(_("-other\t-> %{settings}\n") %{:settings=>@settings['othersettings']}) end
		@settings['log'].append_2_log(_("\nCDDB INFO\n"))
		@settings['log'].append_2_log(_("\nArtist\t= ") + @settings['cd'].artist)
		@settings['log'].append_2_log(_("\nAlbum\t= ") + @settings['cd'].album)
		@settings['log'].append_2_log(_("\nYear\t= ") + @settings['cd'].year)
		@settings['log'].append_2_log(_("\nGenre\t= ") + @settings['cd'].genre)
		@settings['log'].append_2_log(_("\nTracks\t= ") + @settings['cd'].audiotracks.to_s + "\n\n")
		@settings['cd'].audiotracks.times do |track|
			@settings['log'].append_2_log("#{sprintf("%02d", track + 1)} - #{@settings['cd'].tracklist[track]}\n")
		end
		@settings['log'].append_2_log(_("\nSTATUS\n\n"))
	end
	
	def compute_percentage
		@settings['percentages'] = Hash.new() #progress for each track
		totalSectors = 0.0 # It can be that the user doesn't want to rip all tracks, so calculate it
		@settings['tracksToRip'].each{|track| totalSectors += @settings['cd'].length_in_sectors[track - 1]} #update totalSectors
		@settings['tracksToRip'].each{|track| @settings['percentages'][track] = @settings['cd'].length_in_sectors[track-1] / totalSectors}
	end	

	def temp_dir # use a directory named temp where encoding takes place
		@settings['temp_dir'] = File.dirname(@output_dirs[0]) + '/temp/' #Asking the dirname of a directory returns the directory of one higher level
		unless File.directory?(@settings['temp_dir']) : Dir.mkdir(@settings['temp_dir']) end # If the directory not yet exists, create it
	end
end
